/*
	PROforma, PROgs FOnt & Raster MAnager

	QL screen picture driver
	contains :	QPTR _pic driver (mono, mode 4 and mode 8)
				mode 4 screen, 512x256
				mode 8 screen, 256x256
				The PAINTER mode 4 compressed screen
				The PAINTER mode 8 compressed screen

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 19 December 1995
	12.06.1996 pictures now start with FileInfo structure
*/

#include <io_h>
#include "gstate_h"

Error PFPictureElement(Gstate gstate, pt xsiz, pt ysiz, pt xorg, pt yorg);

typedef struct {
	short xsiz,ysiz;	/* size in pixels -(1,1) */
	char data[2];		/* start of actual data */
						/* 2 or 3 LSbits are colour, rest is counter-1 */
} Painter;

static Error Handle(int command,...)
{
	return ERR_NIMP;
}

static Error noRecognize(Gstate gstate, char *base)
{ return ERR_INAM; }

static Error qptrRecognize(Gstate gstate, char *base)
{
	/* try to recognize this as a QPTR picture, else ERR_ITNF */
	FileInfo *fi=(FileInfo *)base;
	WindowArea *area=(WindowArea *)(base+sizeof(FileInfo)-offsetof(WindowArea,flag));

	if (fi->type==FILETYPE_NORMAL &&
		area->flag==WINDOWAREA_FLAG &&
		area->xsiz>0 && area->ysiz>0 &&
		area->line_increment>0 && area->line_increment<max(5,area->xsiz) &&
		( area->mode==WINDOWMODE_2 ||
		  area->mode==WINDOWMODE_M4 ||
		  area->mode==WINDOWMODE_M8 ) &&
		area->zero==0)
		return ERR_OK;
	else
		return ERR_INAM;
}

static Error paint4Recognize(Gstate gstate, char *base)
{
	FileInfo *fi=(FileInfo *)base;

	if (fi->type==6)
		return ERR_OK;
	else
		return ERR_INAM;
}

static Error paint8Recognize(Gstate gstate, char *base)
{
	FileInfo *fi=(FileInfo *)base;

	if (fi->type==11)
		return ERR_OK;
	else
		return ERR_INAM;
}

static Error screenAspectRatio(Gstate gstate, char *base, int *xratio, int *yratio)
{
	*xratio=4; *yratio=3;
	return ERR_OK;
}

static Error qptrAspectRatio(Gstate gstate, char *base, int *xratio, int *yratio)
{
	/* there is no way of knowing this for sure, so guess ! */
	WindowArea *area=(WindowArea *)(base+sizeof(FileInfo)-offsetof(WindowArea,flag));
	*xratio=area->xsiz*4;
	*yratio=area->ysiz*3;
	/* hey, I'm not sure */
	return ERR_ITNF;
}

static Error mode2ColourInfo(Gstate gstate, char *base, int *count, int *space)
{
	*count=2;
	*space=PF_COLOURSPACE_GRAYSHADE;
	return ERR_OK;
}

static Error mode4ColourInfo(Gstate gstate, char *base, int *count, int *space)
{
	*count=4;
	*space=PF_COLOURSPACE_RGB;
	return ERR_OK;
}

static Error mode8ColourInfo(Gstate gstate, char *base, int *count, int *space)
{
	*count=8;
	*space=PF_COLOURSPACE_RGB;
	return ERR_OK;
}

static Error qptrColourInfo(Gstate gstate, char *base, int *count, int *space)
{
	WindowArea *area=(WindowArea *)(base+sizeof(FileInfo)-offsetof(WindowArea,flag));

	switch(area->mode)
	{
	case WINDOWMODE_2:
		return mode2ColourInfo(gstate,base,count,space);
	case WINDOWMODE_M4:
		return mode4ColourInfo(gstate,base,count,space);
	case WINDOWMODE_M8:
		return mode8ColourInfo(gstate,base,count,space);
	}
	return ERR_ISYN;
}

static Error mode2ColourGet(Gstate gstate, char *base, int which, void *colour)
{
	if (which)
		*((pt *)colour)=0;
	else
		*((pt *)colour)=pt_hundred;
	return ERR_OK;
}

static Error mode8ColourGet(Gstate gstate, char *base, int which, void *colour)
{
	ColourRGB *rgb=(ColourRGB *)colour;
	rgb->red=rgb->green=rgb->blue=pt_hundred;
	if (!(which&4))
		rgb->green=0;
	if (!(which&2))
		rgb->red  =0;
	if (!(which&1))
		rgb->blue =0;
	return ERR_OK;
}

static Error mode4ColourGet(Gstate gstate, char *base, int which, void *colour)
{
	ColourRGB *rgb=(ColourRGB *)colour;
	rgb->red=rgb->green=rgb->blue=pt_hundred;
	if (!(which&2))
		rgb->green=0;
	if (!(which&1))
		rgb->red  =0;
	if (which!=3)
		rgb->blue =0;
	return ERR_OK;
}

static Error qptrColourGet(Gstate gstate, char *base, int which, void *colour)
{
	WindowArea *area=(WindowArea *)(base+sizeof(FileInfo)-offsetof(WindowArea,flag));

	switch(area->mode)
	{
	case WINDOWMODE_2:
		return mode2ColourGet(gstate,base,which,colour);
	case WINDOWMODE_M4:
		return mode4ColourGet(gstate,base,which,colour);
	case WINDOWMODE_M8:
		return mode8ColourGet(gstate,base,which,colour);
	}
	return ERR_ISYN;
}

Error scr(Gstate gstate, int mode, char *base, int linc,
		 int xpix, int ypix, pt xsiz, pt ysiz,
		 Error (*ColourSelect)(Gstate, int))
{
	int xpixcount;			/* counter for x */
	int colour, backcolour=0;
	short bit=0x80;
	int length, lastcolour;
	char *buff=base;
	int posx;
	pt posy=0;

	/* the background colour is random (well 0) - actually choosing */
	/* the background colour would require running through the entire */
	/* picture once more -> expensive */

	/* first check whether some part of the picture is visible ! */
	ColourSelect(gstate,backcolour);
	if (PFPictureElement(gstate,xsiz,ysiz,0,0)) return ERR_OK;

	xsiz/=xpix;
	ysiz/=ypix;

	while(ypix)
	{
		posx=0;

		/* see if this line is clipped - and set background colour */
		ColourSelect(gstate,backcolour);
		if (!PFPictureElement(gstate,xsiz*xpix,ysiz,0,posy))
		{
			length=1;
			/* get colour */
			switch(mode)
			{
			case WINDOWMODE_2:
				lastcolour = ((*base)&bit) ? 1 : 0;
				/* skip to next pixel */
				bit>>=1; if (!bit) bit=0x80, base++;
				break;
			case WINDOWMODE_M4:
				lastcolour = ((*base)&bit) ? 2 : 0;
				lastcolour|= ((*(base+1))&bit) ? 1 : 0;
				/* skip to next pixel */
				bit>>=1; if (!bit) bit=0x80, base+=2;
				break;
			case WINDOWMODE_M8:
				lastcolour = ((*base)&bit) ? 4 : 0;
				lastcolour|= ((*(base+1))&bit) ? 2 : 0;
				lastcolour|= ((*(base+1))&(bit>>1)) ? 1 : 0;
				/* skip to next pixel */
				bit>>=2; if (!bit) bit=0x80, base+=2;
				break;
			}
			for(xpixcount=xpix-1; xpixcount; xpixcount--)
			{
				/* get colour */
				switch(mode)
				{
				case WINDOWMODE_2:
					colour = ((*base)&bit) ? 1 : 0;
					/* skip to next pixel */
					bit>>=1; if (!bit) bit=0x80, base++;
					break;
				case WINDOWMODE_M4:
					colour = ((*base)&bit) ? 2 : 0;
					colour|= ((*(base+1))&bit) ? 1 : 0;
					/* skip to next pixel */
					bit>>=1; if (!bit) bit=0x80, base+=2;
					break;
				case WINDOWMODE_M8:
					colour = ((*base)&bit) ? 4 : 0;
					colour|= ((*(base+1))&bit) ? 2 : 0;
					colour|= ((*(base+1))&(bit>>1)) ? 1 : 0;
					/* skip to next pixel */
					bit>>=2; if (!bit) bit=0x80, base+=2;
					break;
				}

				if (colour!=lastcolour)
				{
					if (lastcolour!=backcolour)
					{
						ColourSelect(gstate,lastcolour);
						PFPictureElement(gstate,xsiz*length,ysiz,xsiz*posx,posy);
					}
					/* skip pixels on page */
					posx+=length;
					lastcolour=colour;
					length=1;
				} else
					length++;
			}
			/* there may be a sequence left at the end of the line */
			if (colour!=backcolour)
			{
				ColourSelect(gstate,colour);
				PFPictureElement(gstate,xsiz*length,ysiz,xsiz*posx,posy);
			}
		}
		posy+=ysiz;
		ypix--;
		base=(buff+=linc); bit=0x80;
	}
	return ERR_OK;
}

Error paint(Gstate gstate, short type, char *base, int xpix, int ypix,
			pt xsiz, pt ysiz, Error (*ColourSelect)(Gstate, int))
{
	int xpixcount;					/* counter for x */
	int length=0, drawlen;
	int colour, backcolour=0;
	int posx, posy=0;
	Error nodraw;

	/* the background colour is random (well 0) - actually choosing */
	/* the background colour would require running through the entire */
	/* picture once more -> expensive */

	/* first check whether some part of the picture is visible ! */
	ColourSelect(gstate,backcolour);
	if (PFPictureElement(gstate,xsiz,ysiz,0,0)) return ERR_OK;

	xsiz/=xpix;
	ysiz/=ypix;

	while(ypix)
	{
		posx=0;

		/* see if this line is clipped - and set background colour */
		ColourSelect(gstate,backcolour);
		nodraw=PFPictureElement(gstate,xsiz*xpix,ysiz,0,posy);

		xpixcount=xpix;
		while(xpixcount)
		{
			/* get colour */
			if (!length)
			{
				if (!type)
				{
					colour=*base++;
					length=((colour>>2)&0x3f)+1;
					colour=(colour&3);
				} else {
					colour=*base++;
					length=((colour>>3)&0x1f)+1;
					colour=(colour&7);
				}
			}

			if (length>xpixcount) drawlen=xpixcount;
							 else drawlen=length;

			/* set colour and draw */
			if (!nodraw && colour!=backcolour)
			{
				ColourSelect(gstate,colour);
				PFPictureElement(gstate,xsiz*drawlen,ysiz,xsiz*posx,posy);
			}

			/* skip pixels on page */
			posx+=drawlen;
			length-=drawlen;
			xpixcount-=drawlen;
		}
		posy+=ysiz;
		ypix--;
	}
	return ERR_OK;
}


static Error qptrDisplay(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	WindowArea *area=(WindowArea *)(base+sizeof(FileInfo)-offsetof(WindowArea,flag));

	return scr(gstate,area->mode,area->data,area->line_increment,
			   (area->mode==WINDOWMODE_M8)?(area->xsiz/2):area->xsiz,
			   area->ysiz, xsiz,ysiz, ColourSelect);
}

static Error mode4Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	return scr(gstate,WINDOWMODE_M4,base+sizeof(FileInfo),128,512,256,xsiz,ysiz,ColourSelect);
}

static Error paint4Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	Painter *p=(Painter *)(base+sizeof(FileInfo));
	return paint(gstate,0,p->data,p->xsiz+1,p->ysiz+1,xsiz,ysiz,ColourSelect);
}

static Error mode8Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	return scr(gstate,WINDOWMODE_M8,base+sizeof(FileInfo),128,256,256,xsiz,ysiz,ColourSelect);
}

static Error paint8Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	Painter *p=(Painter *)(base+sizeof(FileInfo));
	return paint(gstate,1,p->data,(p->xsiz+1)/2,p->ysiz+1,xsiz,ysiz,ColourSelect);
}

PICTdriver init5 = {
	NULL, PF_PICTUREDRIVER,
	"The PAINTER, mode 8 compressed picture",

	/* can we recognize the picture */
	paint8Recognize,
	screenAspectRatio,

	/* display the picture */
	paint8Display,

	/* info on colours used in the picture */
	mode8ColourInfo,
	mode8ColourGet,

	Handle
};

PICTdriver init4 = {
	&init5, PF_PICTUREDRIVER,
	"The PAINTER, mode 4 compressed picture",

	/* can we recognize the picture */
	paint4Recognize,
	screenAspectRatio,

	/* display the picture */
	paint4Display,

	/* info on colours used in the picture */
	mode4ColourInfo,
	mode4ColourGet,

	Handle
};

PICTdriver init3 = {
	&init4, PF_PICTUREDRIVER,
	"QPTR Picture",

	/* can we recognize the picture */
	qptrRecognize,
	qptrAspectRatio,

	/* display the picture */
	qptrDisplay,

	/* info on colours used in the picture */
	qptrColourInfo,
	qptrColourGet,

	Handle
};

PICTdriver init2 = {
	&init3, PF_PICTUREDRIVER,
	"QL mode 8 screen, 256x256",

	/* can we recognize the picture */
	noRecognize,
	screenAspectRatio,

	/* display the picture */
	mode8Display,

	/* info on colours used in the picture */
	mode8ColourInfo,
	mode8ColourGet,

	Handle
};

PICTdriver init = {
	&init2, PF_PICTUREDRIVER,
	"QL mode 4 screen, 512x256",

	/* can we recognize the picture */
	noRecognize,
	screenAspectRatio,

	/* display the picture */
	mode4Display,

	/* info on colours used in the picture */
	mode4ColourInfo,
	mode4ColourGet,

	Handle
};
