/*
	PROforma, PROGS FOnt & Raster MAnager

	Epson ESC/P printer driver
	suitable for both 9 and 24 pin printers

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 27 September 1993

	this file is based on the PRoforma v1.0? esc/p external driver
	24.01.1996 adjusted for different driver structure in PROforma v1.10
*/

#include "str_h"
#include "io_h"
#include "driv_escp_h"

/* some macros for generating output */
#define printb(s,b) IOPutRawBytes(chid,s,b,NULL)
#define prints(s)	printb(strlen(s),s)
#define printc(c)	{ char s=c; IOPutRawBytes(chid,1,&s,NULL); }

/* routine to fill in a scan line in the transposed data */
static const char setbit[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
static void transpose(char *data, char *dest, int bit, int skip, int linelen)
{
	char mask=setbit[bit&7];
	long i,j;

	dest+=bit>>3;
	for (j=linelen; j; j--)
	{
		char test=*data++;
		for (i=0; i<8; i++)
		{
			if (test & setbit[i]) *dest|=mask;
			dest+=skip;
		}
	}
}
static void transposeh(char *data, char *dest, int bit, int skip, int linelen)
{
	char mask;
	long i,j;

	if (skip==3)
	{
		dest+=(bit&1)*linelen*24;
		bit/=2;
	}
	else
	{
		dest+=(bit%3)*linelen*8;
		bit/=3;
	}
	mask=setbit[bit&7];
	dest+=bit>>3;
	for (j=linelen; j; j--)
	{
		char test=*data++;
		for (i=0; i<8; i++)
		{
			if (test & setbit[i]) *dest|=mask;
			dest+=skip;
		}
	}
}

/* routine to send a line to the printer */
static void printline(Gstate gstate, char res, char *buff, int count, int y_mult)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int xskip=di->xskip;
	int xcount, i,j;
	char *buff_end;
	char line[20];

	for (i=(di->options&HIGH)?(di->pins==8?3:2):1 ; i ; i--)
	{
		/* remove trailing zeros */
		buff_end=buff+count;
		while ( buff_end>buff && buff_end[-1]==0 ) buff_end--;

		if (xcount=(buff_end-buff+y_mult-1)/y_mult)
		{
			/* initialise graphics */
			prints("\033*");
			printc(res);
			printc((xcount+xskip) & 0xff);
			printc((xcount+xskip) >> 8);

			/* make sure x position is correct */
			for (j=xskip*y_mult; j; j--) printc(0);

			/* print the actual data */
			printb(xcount*y_mult,buff);
		}
		buff+=count;
		printc('\r');
		if (di->pins==24)
		{
			if (i>1) prints("\r\n");
		}
		else if (i>1) prints("\033J\001");	/* skip to next position */
	}
}


Error InitPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int y_mult = (di->pins==24 ? 1 : 3);

	/* Initialize the printer and reset the margins. */
	prints("\033@");
	if (di->options&SETA4) prints("\033C\x46");
	prints("\033P\033l");
	printc(0);
	prints("\r\033Q");
	printc((gstate->allbbox.xsiz+di->xskip) * 10 / gstate->printer->xdpi +2);
	if (di->options&HIGH && di->pins==24)
		/* set 1/360" line spacing */
		if (di->options&NECP5)
			prints("\0343\001");
		else
			prints("\034+\001");

	di->skip_lines=di->options&HIGH?di->yskip:di->yskip*y_mult;
	di->bit=0;
	return ERR_OK;
}

Error ShowPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int line_size = gstate->raster->linc;
	int y_mult = (di->pins==24 ? 3 : 1);
	char *data=(char *)gstate->raster->base;
	char *buff=di->buffer;
	char *buff_end; int i;
	int skip_lines=di->skip_lines;
	char res="\0\1\x27\x03\xff\x28"[gstate->printer->xdpi/60-1];
	int lastbit=di->options&HIGH?(di->pins==8?24:48):di->pins;

	for(i=gstate->pagebbox.ysiz ; i ; i--, data+=line_size)
	{
		/* clear the buffer if necessary */
		if (di->bit==0) MEMClear(line_size*lastbit,di->buffer);

		/* Add this scan line in the buffer (byteline) */
		if (di->options&HIGH)
			transposeh(data,buff,di->bit,y_mult,line_size);
		else
			transpose(data,buff,di->bit,y_mult,line_size);
		di->bit++;

		if (di->bit==lastbit)
		{
			buff_end=buff+line_size*lastbit;
			/* remove trailing zeros */
			while (buff_end>buff && !buff_end[-1]) buff_end--;

			if (buff_end==buff)
			{
				if (di->options&HIGH && di->pins==24)
					skip_lines+=lastbit;
				else
					skip_lines+=24;
			}
			else
			{
				/* Vertical tab to the appropriate position. */
				if (di->options&HIGH && di->pins==24)
				{
					if (skip_lines&1) printc('\n');
					skip_lines/=2;
				}
				while (skip_lines>255)
				{
					prints("\033J\377");
					skip_lines-=255;
				}
				if (skip_lines)
				{
					prints("\033J");
					printc(skip_lines);
				}
				printline(gstate, res, buff, line_size*di->pins, y_mult);
				skip_lines=24;
				if (di->options&HIGH)
				{
					if (di->pins==24)
						skip_lines=47;
					else
						skip_lines=22;
				}
			}
			di->bit=0;
		}
	}
	di->skip_lines=skip_lines;
	return ERR_OK;
}

Error StopPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int line_size = gstate->raster->linc;
	int y_mult = (di->pins==24 ? 3 : 1);
	char *data=(char *)gstate->raster->base;
	char *buff=di->buffer;
	char *buff_end; int i;
	int skip_lines=di->skip_lines;
	char res="\0\1\x27\x03\xff\x28"[gstate->printer->xdpi/60-1];
	int lastbit=di->options&HIGH?(di->pins==8?24:48):di->pins;

	if (di->bit==lastbit)
	{
		buff_end=buff+line_size*lastbit;
		/* remove trailing zeros */
		while (buff_end>buff && !buff_end[-1]) buff_end--;

		if (buff_end==buff)
		{
			if (di->options&HIGH && di->pins==24)
				skip_lines+=lastbit;
			else
				skip_lines+=24;
		}
		else
		{
			/* Vertical tab to the appropriate position. */
			if (di->options&HIGH && di->pins==24)
			{
				if (skip_lines&1) printc('n');
				skip_lines/=2;
			}
			while (skip_lines>255)
			{
				prints("\033J\377");
				skip_lines-=255;
			}
			if (skip_lines)
			{
				prints("\033J");
				printc(skip_lines);
			}
			printline(gstate, res, buff, line_size*di->pins, y_mult);
		}
		di->bit=0;
	}
	prints("\f\x1b@");
	return ERR_OK;
}
