/*
	PROforma, PROgs FOnt & Raster MAnager

	FastPattern drawing routines for mode 4 bitmap driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 2 October 1995
	16.12.1995 moved DrawMask routine to different file (inc. clipping)
*/

#include "gstate_h"

char RotateChar(short amount, char what);

Error FastPatternPixel(Bitmap bitmap, short x, short y)
{
	unsigned char *base=bitmap->base;	/* base of bitmap */
	unsigned char *pattern; 		/* FastPattern to use as colour */
	char xb;						/* bit pattern for x */

	/* clip pixel to useable area */
	if (x<0 || x>=bitmap->clip.xsiz ||
		y<0 || y>=bitmap->clip.ysiz)
		return ERR_OK;
	x+=bitmap->clip.xorg;			/* add origin in bitmap !! */
	y+=bitmap->clip.yorg;

	xb=~( 0x80 >> (x&(short)7) );	/* mask to clear the pixel */
	base+=y*bitmap->linc+((x>>2)&(-2)); /* address in bitmap to set pixels */

	pattern=&bitmap->pattern[(y&7)*2];

	*base  &=xb;
	*base++|=*pattern++&~xb;
	*base  &=xb;
	*base  |=*pattern  &~xb;
	return ERR_OK;
}

Error FastPatternLineHorizontal(Bitmap bitmap, short y, short xs, short xe)
{
	unsigned char *base=bitmap->base;	/* base of bitmap */
	unsigned char *pattern; 		/* FastPattern to use as colour */
	char xl, xr;					/* bit pattern for x */
	short dist; 					/* number of pixels to fill */

	/* clip line to useable area */
	if (xe<0 || xs>=bitmap->clip.xsiz ||
		 y<0 ||  y>=bitmap->clip.ysiz)
		return ERR_OK;
	if (xs<0) xs=0;
	if (xe>=bitmap->clip.xsiz) xe=bitmap->clip.xsiz;
	xs+=bitmap->clip.xorg;			/* add origin in bitmap !! */
	xe+=bitmap->clip.xorg;
	y +=bitmap->clip.yorg;
										
	dist=xe-xs;
	if (dist<=0) return ERR_OK; 	/* we need a positive length */

	xl=~(0xff >> (xs&7));			/* mask for pixels at left */
	xr= (0xff >> (xe&7));			/* mask for pixels at right */
	base+=y*bitmap->linc+((xs>>2)&(-2));/* address in bitmap to set pixels */

	pattern=&bitmap->pattern[(y&7)*2];

	dist-=8-(xs&7); 				/* no of pixels to go after first byte */

	if (dist<0) 					/* only one byte to adjust */
	{
		xl|=xr;
		*base  &=xl;
		*base++|=*pattern++&~xl;
		*base  &=xl;
		*base  |=*pattern  &~xl;
	} else {                        /* adjust at least two bytes */
		*base  &=xl;
		*base++|=pattern[0]&~xl;
		*base  &=xl;
		*base++|=pattern[1]&~xl;
		for (dist>>=3; dist; dist--) *((short *)base)++=*((short *)pattern);
		*base  &=xr;
		*base++|=*pattern++&~xr;
		*base  &=xr;
		*base  |=*pattern  &~xr;
	}
	return ERR_OK;
}

Error FastPatternChangeOrigin(Bitmap bitmap, short x, short y)
{
	int xrotate=x-bitmap->patternbox.xorg;
	int yrotate=y-bitmap->patternbox.yorg;
	int i;

	for (i=15 ; i>=0 ; i--)
		bitmap->pattern[i]=RotateChar(xrotate,bitmap->pattern[i]);

	if (yrotate<0) yrotate+=8;

	for ( ; yrotate ; yrotate--)
	{
		short *base=(short *)bitmap->pattern;
		short keep=base[7];

		for (i=6 ; i>=0 ; i--)
			base[i+1]=base[i];
		base[0]=keep;
	}
	bitmap->patternbox.xorg=x;
	bitmap->patternbox.yorg=y;

	return ERR_OK;
}
