/* PROforma, PROgs FOnt & Raster MAnager

	BITMdriver definition file for screen driver

   (c) PROGS, Professional & Graphical Software
   written by Joachim Van der Auwera
   started 28 September 1995
   */

#include "gstate_h"

static Error nimp()    { return ERR_NIMP; }
static Error nothing() { return ERR_OK;   }

Error _LineVertical(Bitmap, short atx, short fromy, short toy);
Error _Block(Bitmap, short xsiz, short ysiz, short xorg, short yorg);
Error _TrapezoidHorizontal(Bitmap, short starty, short endy,
				pt leftstartx, pt leftincr, pt rightstartx, pt rightincr);
Error _TrapezoidVertical(Bitmap, short startx, short endx,
				pt topstarty, pt topincr, pt bottomstarty, pt bottomincr);
Error _HairLine(Bitmap, pt fromx, pt fromy, pt tox, pt toy);

Error BlackPixel(Bitmap, short x, short y);
Error BlackLineHorizontal(Bitmap, short aty, short fromx, short tox);
Error BlackDrawMask(Bitmap, Mask *mask, short x, short y);

Error WhitePixel(Bitmap, short x, short y);
Error WhiteLineHorizontal(Bitmap, short aty, short fromx, short tox);
Error WhiteDrawMask(Bitmap, Mask *mask, short x, short y);

Error FastPatternPixel(Bitmap, short x, short y);
Error FastPatternLineHorizontal(Bitmap, short aty, short fromx, short tox);
Error FastPatternDrawMask(Bitmap, Mask *mask, short x, short y);
Error FastPatternChangeOrigin(Bitmap, short x, short y);

Error PatternPixel(Bitmap, short x, short y);
Error PatternLineHorizontal(Bitmap, short aty, short fromx, short tox);
Error PatternDrawMask(Bitmap, Mask *mask, short x, short y);
Error PatternChangeOrigin(Bitmap, short x, short y);

Error CursorSet(Bitmap, short x, short y);
Error CursorMove(Bitmap, short x, short y);
Error CursorGet(Bitmap, short *x, short *y);

Error PixelRead(Bitmap, int *colour);
Error PixelReadNextX(Bitmap, int *colour);
Error PixelReadPrevX(Bitmap, int *colour);
Error PixelReadNextY(Bitmap, int *colour);
Error PixelReadPrevY(Bitmap, int *colour);
Error PixelSet(Bitmap, int colour);
Error PixelSetNextX(Bitmap, int colour);
Error PixelSetPrevX(Bitmap, int colour);
Error PixelSetNextY(Bitmap, int colour);
Error PixelSetPrevY(Bitmap, int colour);

Error CopyArea(Bitmap from, Bitmap to, short xsiz, short ysiz,
			   short fromx, short fromy, short tox, short toy);


Error BufferSize(short xsiz, short ysiz, Size *size)
{
	*size=((xsiz+7)&-8)*2 * ysiz / 8;
	return ERR_OK;
}


BITMdriver init={
	NULL, PF_BITMAPDRIVER,
	{
		2, 1,					/* 2 planes, each 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		1,						/* altenative bytes in each plane */
	},

	/* drawing routines */
	{                               /* Black drawing routines */
		BlackPixel,
		BlackLineHorizontal,
		_LineVertical,
		_Block,
		_TrapezoidHorizontal,
		_TrapezoidVertical,
		_HairLine,
		BlackDrawMask,
		nothing
	},
	{                               /* White drawing routines */
		WhitePixel,
		WhiteLineHorizontal,
		_LineVertical,
		_Block,
		_TrapezoidHorizontal,
		_TrapezoidVertical,
		_HairLine,
		WhiteDrawMask,
		nothing
	},
	{                               /* FastPattern drawing routines */
		FastPatternPixel,
		FastPatternLineHorizontal,
		_LineVertical,
		_Block,
		_TrapezoidHorizontal,
		_TrapezoidVertical,
		_HairLine,
		FastPatternDrawMask,
		FastPatternChangeOrigin
	},
	{                               /* Pattern drawing routines */
		BlackPixel,
		BlackLineHorizontal,
		_LineVertical,
		_Block,
		_TrapezoidHorizontal,
		_TrapezoidVertical,
		_HairLine,
		BlackDrawMask,
		nimp
	},

	CursorSet,
	CursorMove,
	CursorGet,

	PixelRead,
	PixelReadNextX,
	PixelReadPrevX,
	PixelReadNextY,
	PixelReadPrevY,
	PixelSet,
	PixelSetNextX,
	PixelSetPrevX,
	PixelSetNextY,
	PixelSetPrevY,

	CopyArea,

	BufferSize, /* BufferSize */
	8,8,		/* size of FastPattern */
	BITMAP_DIRECTION_VERTICAL,

	nimp
	};
