/*
	PROforma, PROGS FOnt & Raster MAnager

	Printer drivers for all PCL printers like
	HP DeskJet and variants
	HP LaserJet and variants

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 18 September 1993

	this file is based on HP LaserJet/DeskJet driver for LINEdesign v1
	24.03.1994 adjusted for external drivers
	04.10.1995 for PROforma v1.10 - different driver structure
	29.06.1996 fixed a bug in CompressMode2
*/

#include "io_h"
#include "str_h"

#include "driv_pcl_h"

/* Forward references */
static int CompressMode2(char *, char *, char *);
static int CompressMode3(int, char *, char *, char *);

/* some macros for generating output */
#define printb(s,b) IOPutRawBytes(chid,s,b,NULL)
#define prints(s)	printb(strlen(s),s)

/* set the NrOfCopies at which the page should be printed */
Error SetCopies(Gstate gstate, int count)
{
	gstate->PrinterExtra->copies=count;
}

/* initialise the printer */

Error InitPage(Gstate gstate)
{
	Channel chid=gstate->PrinterExtra->channel;
	DriverInfo *di=gstate->PrinterExtra;
	int type=di->type;
	char line[60];

	if (!(type&NORESET)) prints("\033E");

	/* set number of copies */
	STRFormatPut(line,"\033&l%dX\033*rB",di->copies);
	prints(line);

	/* set raster graphics resolution */
	STRFormatPut(line,"\033*t%dR\033*r0f0A",gstate->printer->xdpi);
	prints(line);
	/* move to top left of page & set current position */
	if (type&MINIMARGIN)
		prints("\033&l-60u120z0E");
	else
		prints("\033&l0e-185U");
	/* select unit and the most compressed mode available */
	di->compression=type&COMPRESS;
	STRFormatPut(line,"\033&u%dD\033*p%dx%dY\033*b%dM",
					gstate->printer->xdpi,di->xskip,di->yskip,di->compression);
	prints(line);
	di->skip_lines=0;		  /* no lines to skip yet */
	if ((di->type&COMPRESS)==LEVEL5) MEMClear(gstate->raster->linc,di->prev_line);
	return ERR_OK;
}


/* Send the buffer to the printer. Use as much compression as possible to */
/* speed up the data transfer. */

Error ShowPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int line_size=gstate->raster->linc;
	char *data=(char *)gstate->raster->base;
	char *end_data;
	char *out_line=di->out_line;
	char *prev_line=di->prev_line;
	char *alt_line=di->alt_line;
	char *out_data;
	int out_len, i;
	int skip_lines=di->skip_lines;
	int pcllevel=di->type&COMPRESS;
	int compression=di->compression;
	char line[40];

	for (i=gstate->pagebbox.ysiz ; i ; i--, data+=line_size )
	{
		end_data=data+line_size;
		/* remove trailing zeros - printing "nothing" is useless ! */
		while (end_data>data && !end_data[-1]) end_data--;
		if (end_data==data)
		{
			if (!skip_lines && pcllevel==LEVEL5)
				MEMClear(line_size,prev_line);
			skip_lines++;
			continue;
		}
		switch (pcllevel)
		{
		case LEVEL5:
			{
				/* use method which compresses most, and switch when useful */
				int len3=CompressMode3(line_size, data, prev_line, out_line);
				int len2=CompressMode2(data, end_data, alt_line);
				MEMCopy(line_size,data,prev_line);
				if (compression==3)
				{
					out_data=out_line;
					out_len=len3;
					if (len2+5<len3)
					{
						compression=2;
						prints("\033*b2M");
						out_data=alt_line;
						out_len=len2;
					}
				}
				else
				if (compression==2)
				{
					out_data=alt_line;
					out_len=len2;
					if (len3+5<len2)
					{
						compression=3;
						prints("\033*b3M");
						out_data=out_line;
						out_len=len3;
					}
				}
			}
			break;
		case LEVEL4:
			out_data=out_line;
			out_len=CompressMode2(data, end_data, out_line);
			break;
		case LEVEL3:
			out_data=data;
			out_len=end_data - data;
			break;
		}
		/* Skip blank lines if any */
		if (skip_lines)
		{
			/* move down from current position */
			if (di->type&UNIT)
				STRFormatPut(line,"\033*p+%dY",skip_lines);
			else
				STRFormatPut(line,"\033*p+%dY",skip_lines*300/gstate->printer->xdpi);
			prints(line);
			skip_lines=0;
		}
		/* transfer raster graphics */
		STRFormatPut(line,"\033*b%dW",out_len);
		prints(line);
		/* send the row */
		printb(out_len,out_data);
	}
	di->skip_lines=skip_lines;
	di->compression=compression;
	return ERR_OK;
}

/* handles everything to end the printing */

Error StopPage(Gstate gstate)
{
	Channel chid=gstate->PrinterExtra->channel;

	/* end raster graphics & eject page */
	prints("\033*rB\033&l0H");
	return ERR_OK;
}

/* mode 2 compression routine or TIFF (rev. 4.0) encoding */
/* variation on run length encoding */
/* theoretically, the output can be at most twice as long as the input */
/* however - the CompressMode2 routine produces an output with maximum */
/* length size+size/127+1 */

static int CompressMode2(char *line, char *end, char *output)
{
	char *org_output=output;
	char *run_start;
	char previous;
	char identical=FALSE;
	int counter;

	while (line<end)
	{
		run_start=line;
		if (identical)
		{
			previous=*line++;
			while (line<end && *line==previous) line++;
			for (counter=line-run_start ; counter>128 ; counter-=128)
			{
				*output++=-127;
				*output++=previous;
			}
			if (counter>1)
			{
				*output++=-counter+1;
				*output++=previous;
			}
			else
				line--;
			identical=FALSE;
		}
		else
		{
			/* need three identical bytes for compression to stay optimal */
			/* I could also request another identical byte, this would */
			/*	- make longer runs - might be more efficient for the printer */
			/*	  but then again, sending data to the printer is so slow */
			/*	- longer runs makes it easier to get a break because you can */
			/*	  run out of counter bits, therefore, the compression can */
			/*	  be BETTER by only going for three identical bytes */
			while (line+2<end && (line[1]!=line[0] || line[2]!=line[0])) line++;
			if (line+2>=end) line=end;
			for (counter=line-run_start ; counter>128 ; counter-=128)
			{
				*output++=127;
				MEMCopy(128,run_start,output);
				output+=128;
				run_start+=128;
			}
			if (counter)
			{
				*output++=counter-1;
				MEMCopy(counter,run_start,output);
				output+=counter;
			}
			identical=TRUE;
		}
	}
	return output-org_output;
}

/* mode 3 compression routine or delta row compression */
/* similar in spirit to mode2 compression, but instead of bytes */
/* identical to the previous (horizontal), it uses bytes identical to */
/* the byte just above */
/* the output can have a maximum length of size+size/8+1 */

static CompressMode3(int size, char *line, char *prev, char *output)
{
	char *org_output=output;
	char *run_start;
	int counter,i;

	size--;
	while (size)
	{
		/* skip identical bytes */
		run_start=line;
		while (*line++==*prev++ && size) size--;
		if (size)
		{
			size--;
			counter=line-run_start-1;
			run_start=line-1;
			for (i=1 ;
				 i<8 && *line!=*prev && size ;
				 line++, prev++, i++, size--);
			if (counter<31)
				*output++=((i-1)<<5)+counter;
			else
			{
				*output++=((i-1)<<5)+31;
				for (counter-=31 ; counter>254 ; counter-=255)
					*output++=255;
				*output++=counter;
			}
			MEMCopy(i,run_start,output);
			output+=i;
		}
	}
	return output-org_output;
}
